(in-package :max)
(defclass function-type (max-type)
        nil
)
(defparameter *functions* (make-type :class 'function-type :key 'FUNCTION))
(add-init-hook #'(lambda () (add-name *functions* "functions")))
(defclass function-class ()
        ((mapping :accessor mapping
                :initarg :mapping
        ))
)
(defun function-implementation (funcval)
        (third (addr-value funcval))
)

(defclass enumerated-function (function-class)
        ; Use accessor and initarg from parent for this slot
        ((mapping :type hash-table))
)
(defmethod addr-value-equal ((f enumerated-function) a b)
        (eq a b)
)
(defmethod apply-address ((func enumerated-function) param)
    (make-address-form
        (addr-type param)
        (gethash (addr-value param) (mapping func))
    )
)
(defun function-change-helper (h x)
        (setf (gethash (addr-value (symbol-value (car x))) h)
                (addr-value (symbol-value (cdr x))))
)
(defun function-values (&rest args)
        (let ((h (make-hash-table)))
                (map
                        nil
                        #'(lambda (x)
                                (function-change-helper h x))
                        args
                )
                h
        )
)
(defmethod function-change ((func enumerated-function) tup)
        (let ((h (mapping func)))
                (function-change-helper h tup)
        )
)
(defmethod function-change-2 ((func enumerated-function) arg result)
        (let ((h (mapping func)))
                (setf (gethash (addr-value arg) h)
                        (addr-value result))
        )
)
; Apply a function to param and return the address of the result.
; Wrap apply-address, a generic function
(defun function-apply (funcaddr param)
    (apply-address
        (function-implementation funcaddr) 
        param      ; assume domain is prechecked
    )
)

(defun function-quotient (func_addr result_addr)
        (let ((quotient (get-empty-set))
              (divisor (addr-value result_addr)) 
              (dom (function-domain func_addr)))

                (maphash
                        #'(lambda (key val)
                                (if (addr-value-equal dom divisor val)
                                        (setf quotient (set-add quotient
                                                (make-address-form dom key)
                                        ))
                                )
                        )

                        (mapping (function-implementation func_addr))
                )

                quotient
        )
)
(add-init-hook #'(lambda ()
    (let ((function-quotient
            (make-function 'builtin-function *functions* *functions*
                #'(lambda (func)
                    (make-function 'builtin-function 
                        (function-codomain func) *set-type*
                        #'(lambda (divisor)
                            (function-quotient func divisor))
                    ))
            )))

        (add-public-function function-quotient)
        (add-name function-quotient "quotient")
    )
))
(defun make-function (class domain codomain mapping &optional key)
        (let* ((obj (make-instance class :mapping mapping))
                (addr (make-address-form (addr-value *functions*) (list (addr-value domain) (addr-value codomain) obj))))

                addr
        )
)
(defclass builtin-function (function-class)
        ((mapping :type function)) ;store a lambda expression
)
(defmethod apply-address ((bf builtin-function) arg)
        (apply (mapping bf) (list arg))
)

(defun function-domain (func)
    (make-address-form (addr-value *types*) (first (addr-value func)))
)

(defun function-codomain (func)
    (make-address-form (addr-value *types*) (second (addr-value func)))
)
(defparameter *domain* (make-function 'builtin-function 
    *functions*
    *types*
    #'(lambda (func)
       (function-domain func)
    )
))
(defparameter *codomain* (make-function 'builtin-function
    *functions*
    *types*
    #'(lambda (func) (function-codomain func))
))
(add-init-hook #'(lambda ()
    (add-public-function *domain*)
    (add-public-function *codomain*)
    (add-name *domain* "domain")
    (add-name *codomain* "codomain")
))
(defun identity-function (type)
        (make-function 'builtin-function type type
                #'(lambda (addr) addr))
)


