(in-package max)
; The Pointers
(defparameter *ptrs* (make-type))
; Follow a pointer
(defun deref-ptr (ptr)
        ; return the address pointed at
        (nth-value 0 (gethash (addr-value ptr) (type-hash *ptrs*)))
)

; Make a pointer that refers to its own address,
; i.e. self-referencing.
(defun make-self-ptr ()

    ; Get address for the pointer
    (let* ((addr (make-address *ptrs*)) (key (addr-value addr)))

        ; Point the pointer at itself
        (setf (gethash key (type-hash *ptrs*)) addr)

        ; Return the address of the new self-referencing pointer
        (make-address-form (addr-value *ptrs*) key)

    )
)
(defun make-ptr (addr &key restricted)
        (make-address *ptrs* :value addr)
)
(defun change-ptr (ptr addr)

    ; Get the address ptr currently points at
    (let ((oldaddr (deref-ptr ptr)))

        ; Type-check if this pointer is restricted
        (cond
                ((and
                        (eq (addr-type ptr) 'restptr)
                        (not
                                (eq (addr-type oldaddr) (addr-type addr))
                        )
                 )

                        (format t "Type mismatch~%")

                        (return-from change-ptr nil)
                )

        )

        ; If ptr already points at addr, we're done
        ; (typecheck above should be done first, or else
        ;  it would appear that it works to set a restricted pointer
        ;  that was already pointing at what we want, even if the
        ;  types didn't match)
        (if (eq addr oldaddr) nil)

        ; update the ptr
        (setf (gethash (addr-value ptr) (type-hash *ptrs*)) addr)

        ; return nothing
        nil
    )
)
