(in-package max)
(defclass set-type (max-type)
        nil
)
(defparameter *set-type* (make-type :class 'set-type :key 'SET))
(add-init-hook #'(lambda () (add-name *set-type* "sets")))
(defmethod addr-value-equal ((s set-type) a b)
        (set-equal a b)
)
; Create a set address with hash table h
(defun make-set-helper (type h)
        (make-address-form (addr-value *set-type*) (cons (addr-value type) h))
)
(defun set-implementation (setaddr)
        (cdr (addr-value setaddr))
)
(defun set-type (setaddr)
        (make-address-form (addr-value *types*) (car (addr-value setaddr)))
)
(defun get-empty-set (&optional (type *objects*))
        ; return the empty-set address
        (make-set-helper type (make-hash-table :test 'equalp))
)
(defparameter get-empty-set (make-function 'builtin-function *null-type* *set-type* 
        #'(lambda (ignore-nil) (get-empty-set)) 'GET-EMPTY-SET
))
(add-init-hook #'(lambda ()
    (add-public-function get-empty-set)
    (add-name get-empty-set "emptyset")
))
(defparameter get-empty-typed-set (make-function 'builtin-function *types* *set-type*
        #'(lambda (type) (get-empty-set type)) 'GET-EMPTY-TYPED-SET
))
(add-init-hook #'(lambda ()
    (add-public-function get-empty-typed-set)
    (add-name get-empty-typed-set "emptyset-typed")
))
(defun set-add (set_addr addr)

        ; Create a copy of the original hash table
        (let ((newhash (copy-hash (set-implementation set_addr))))

                ; Add the addr to the newhash
                (setf (gethash (addr-value addr) newhash) nil)

                ; return a new set address
                (make-set-helper (set-type set_addr) newhash)
        )
)
(defparameter set-add
            (make-function 'builtin-function
                *set-type*
                *functions*
                #'(lambda (set_addr)
                    (make-function 'builtin-function
                        (set-type set_addr)
                        *set-type*
                        #'(lambda (value_to_add)
                            (set-add set_addr value_to_add)
                        )
                    )
                )
            )
)

(add-init-hook #'(lambda ()
        (add-public-function set-add)
        (add-name set-add "add")
    )
)

(defun set-member (set addr)
        (hash-exists (addr-value addr) (set-implementation set))
)
(add-init-hook #'(lambda ()
    (let ((set-member
            (make-function 'builtin-function
                *set-type*
                *functions*
                #'(lambda (set_addr)
                    (make-function 'builtin-function
                        (set-type set_addr)
                        *boolean-type*
                        #'(lambda (value_to_check)
                            (set-member set_addr value_to_check)
                        )
                    )
                )
            )))

        (add-public-function set-member)
        (add-name set-member "member")
    )
))

(defun mapset (func set)
        (maphash 
                #'(lambda (item ignore-hashval)
                        (funcall func (make-address-form (addr-value (set-type set)) item))
                )

                (set-implementation set)
        )
)
(add-init-hook #'(lambda ()
    (let ((mapset
            (make-function 'builtin-function
                *set-type*
                *functions*
                #'(lambda (set_addr)
                    (make-function 'builtin-function
                        *functions*
                        *set-type*
                        #'(lambda (func_addr)
                          (let ((result (get-empty-set (set-type set_addr))))
                            (mapset
                                #'(lambda (item) 
                                    (setq result
                                        (set-add result 
                                            (function-apply func_addr item))))
                                 set_addr
                            )
                            result
                        ))
                ))
        )))

        (add-public-function mapset)
        (add-name mapset "map")
    )
))

(defun set-union (a b)
        (let ((c (get-empty-set)))
                (mapset #'(lambda (item) (setq c (set-add c item))) a)
                (mapset #'(lambda (item) (setq c (set-add c item))) b)

                ; return the new set
                c
        )
)
(add-init-hook #'(lambda ()
    (let ((set-union
            (make-function 'builtin-function
                (type-product *set-type* *set-type*)
                *set-type*
                #'(lambda (twosets)
                    (apply #'set-union (product-values twosets))
            ))
        ))

        (add-public-function set-union)
        (add-name set-union "union")
    )
))
(defun set-intersect (a b)
        (let ((c (get-empty-set)))
                (mapset #'(lambda (item)
                                (if (set-member b item)
                                        (setq c (set-add c item))
                                )
                        )

                        a
                )

                ; return the new set
                c
        )
)
(add-init-hook #'(lambda ()
    (let ((set-intersect
            (make-function 'builtin-function
                (type-product *set-type* *set-type*)
                *set-type*
                #'(lambda (twosets)
                    (apply #'set-intersect (product-values twosets))
            ))
        ))
        (add-public-function set-intersect)
        (add-name set-intersect "intersect")
    )
))

(defun set-diff (a b)
        (let ((c (get-empty-set)))
                (mapset #'(lambda (item)
                                (if (not (set-member b item))
                                        (setq c (set-add c item))
                                )
                        )

                        a
                )

                ; return the new set
                c
        )
)
(add-init-hook #'(lambda ()
    (let ((set-diff
            (make-function 'builtin-function
                (type-product *set-type* *set-type*)
                *set-type*
                #'(lambda (twosets)
                    (apply #'set-diff (product-values twosets))
            ))
        ))
        (add-public-function set-diff)
        (add-name set-diff "difference")
    )
))
(defun set-remove (origset item)
        (let ((newset (copy-hash (set-implementation origset))))
                (remhash item newset)

                ; return the new set
                (make-set-helper (set-type origset) newset)
        )
)
(add-init-hook #'(lambda ()
    (let ((set-remove
            (make-function 'builtin-function
                *set-type*
                *functions*
                #'(lambda (set_addr)
                    (make-function 'builtin-function
                        (set-type set_addr)
                        *set-type*
                        #'(lambda (item_to_delete)
                            (set-remove set_addr item_to_delete)
                    ))
            ))
        ))

        (add-public-function set-remove)
        (add-name set-remove "remove")
    )
))

(defun set-subset (a b)
        (mapset #'(lambda (item)
                        (if (not (set-member b item))
                                (return-from set-subset nil)
                        )
                )
                a
        )
        t
)
(add-init-hook #'(lambda ()
    (let ((set-subset
            (make-function 'builtin-function
                (type-product *set-type* *set-type*)
                *boolean-type*
                #'(lambda (twosets)
                    (apply #'set-subset (product-values twosets))
            ))
        ))

        (add-public-function set-subset)
        (add-name set-subset "subset")
    )               
))

(defun set-equal (a b)
        (and (set-subset a b) (set-subset b a)) )

(defun set-quotient (set func result)
    (let ((newset (get-empty-set (set-type set))))
        (mapset
            #'(lambda (item)
                (if (addr-equal (function-apply func item) result)
                    (setq newset (set-add newset item))
                )
            )
            set
        )
        newset
    )
)
; first 3-argument function
(add-init-hook #'(lambda ()
    (let ((set-quotient
            (make-function 'builtin-function
                (type-product *set-type* *functions* *objects*)
                *set-type*
                #'(lambda (sfr) (apply #'set-quotient (product-values sfr)))
            )))

        (add-public-function set-quotient)
        (add-name set-quotient "quotient")
    )
))

(defun set-count (setaddr)
    (hash-table-count (set-implementation setaddr))
)

(defun select (relation predicate)
    (set-quotient relation predicate TRUE)
)
(add-init-hook #'(lambda ()
    (let ((select
            (make-function 'builtin-function
                (type-product *set-type* *functions*)
                *set-type*
                #'(lambda (set_and_func) 
                    (apply #'select (product-values set_and_func))
                )
            )))

        (add-public-function select)
        (add-name select "select")
    )
))

(defun update (relation func)
    (let ((newset (get-empty-set (set-type relation))))
        (mapset
            #'(lambda (item)
                (setq newset (set-add newset (function-apply func item)))
            )
            relation
        )
        newset
    )
)

(add-init-hook #'(lambda ()
    (let ((update
            (make-function 'builtin-function
                (type-product *set-type* *functions*)
                *set-type*
                #'(lambda (set_and_func)
                    (apply #'update (product-values set_and_func))
            ))))

        (add-public-function update)
        (add-name update "update")
    )
))

(defun relation-delete (relation predicate)
    (set-quotient relation predicate FALSE)
)
(add-init-hook #'(lambda ()
    (let ((delete
            (make-function 'builtin-function
                (type-product *set-type* *functions*)
                *set-type*
                #'(lambda (set_and_func)
                    (apply #'delete (product-values set_and_func))
            ))))

        (add-public-function delete)
        (add-name delete "delete")
    )
))
