; Concatenator: a widget to display a function component for the
; string "concatenate" function.  It's an upside-down triangle which
; concatenates 2 strings (left and right).  The inputs connect to the
; top corners and the bottom corner outputs the result.

(in-package :sync)

(defparameter *concatenator-type* (maxdb:make-type *db* :id 'concatenator))
(defparameter *concatenator-inputl* (maxdb:make-function *db* *concatenator-type* *sink-type* :id 'concatenator-inputl))
(defparameter *concatenator-inputr* (maxdb:make-function *db* *concatenator-type* *sink-type* :id 'concatenator-inputr))
(defparameter *concatenator-output* (maxdb:make-function *db* *concatenator-type* *source-type* :id 'concatenator-output))
(defparameter *concatenator-scene* (maxdb:make-function *db* *concatenator-type* *scene-type* :id 'concatenator-scene))
(defparameter *concatenator-x* (maxdb:make-function *db* *concatenator-type* maxdb:*number-type* :id 'concatenator-x)) 
(defparameter *concatenator-y* (maxdb:make-function *db* *concatenator-type* maxdb:*number-type* :id 'concatenator-y)) 
(defparameter *concatenator-z* (maxdb:make-function *db* *concatenator-type* maxdb:*number-type* :id 'concatenator-z)) 

(defclass glcat (globj glpointable)
  ((id :initform nil :initarg :glcat-id :accessor glcat-id)
   (vminputl :initform nil :initarg :inputl :accessor glcat-inputl)
   (vminputr :initform nil :initarg :inputr :accessor glcat-inputr)
   (vmoutput :initform nil :initarg :output :accessor glcat-output)))

(defun glcat-callback-left (sink valuel)
  (let* ((parent (sink-parent sink))
         (valuer (sink-value (glcat-inputr parent))))
    (setf (sink-value sink) valuel)
    (unless (null valuer)
      (maxvmm:update (glcat-output parent) (concatenate 'string valuel valuer)))))

(defun glcat-callback-right (sink valuer)
  (let* ((parent (sink-parent sink))
         (valuel (sink-value (glcat-inputl parent))))
    (setf (sink-value sink) valuer)
    (unless (null valuel)
      (maxvmm:update (glcat-output parent) (concatenate 'string valuel valuer)))))

(defmethod draw ((obj glcat) width height)
  (call-next-method)
  (gl:translate-f (obj-x obj) (obj-y obj) (obj-z obj))
  (gl:begin gl:+triangles+)
  (mapc #'(lambda (coords) (apply #'gl:vertex-3f coords)) '((0f0 -0.2f0 0f0) (0.2f0 0.2f0 0f0) (-0.2f0 0.2f0 0f0)))
  (gl:end)
  (if (selected obj)
      (progn 
        (gl:color-3f 1f0 1f0 1f0)
        (gl:begin gl:+line-loop+)
             (mapc #'(lambda (coords) (apply #'gl:vertex-3f coords)) '((0f0 -0.2f0 0f0) (0.2f0 0.2f0 0f0) (-0.2f0 0.2f0 0f0)))
             (gl:end))))

(defun make-cat-brush (font)
  #'(lambda (x y)
      (multiple-value-bind (objx objy objz)
          (position-object x y)

        (format t "cat brush: obj=<~A,~A,~A>~%"
                objx objy objz)

        (make-cat objx objy objz))))

(defmethod initialize-instance :after ((cobj glcat) &rest initargs)
  (setf (sink-parent (glcat-inputl cobj)) cobj)
  (setf (sink-parent (glcat-inputr cobj)) cobj))

(defun make-cat (x y z)
  (let* ((id (+ 1 (maxdb:lastid *db* 'concatenator)))
         (idsinkl (+ 1 (maxdb:lastid *db* 'sink)))
         (idsinkr (+ 1 idsinkl))
         (idsource (+ 1 (maxdb:lastid *db* 'source)))
         (mysource (make-instance 'glsource :id idsource :type 'string))
         (mysinkl (make-instance 'funcom-sink :id idsinkl :type 'string
                                 :callback #'glcat-callback-left))
         (mysinkr (make-instance 'funcom-sink :id idsinkr :type 'string
                                 :callback #'glcat-callback-right))
         (instance (make-instance 'glcat :glcat-id id
                                  :inputl mysinkl :inputr mysinkr :output mysource
                                  :color (make-color :r 0f0 :g 1f0 :b 0f0 :a 1f0))))

    (maxdb:type-add-value *concatenator-type* :id id)
    (maxdb:function-update *concatenator-scene* id (max:make-number (scene-id *scene*)))
    (maxdb:function-update *concatenator-x* id (max:make-number x))
    (maxdb:function-update *concatenator-y* id (max:make-number y))
    (maxdb:function-update *concatenator-z* id (max:make-number z))

    (maxdb:type-add-value *sink-type* :id idsinkl)
    ;XXX- API to create new max type needs improved and used here.
    (maxdb:function-update *concatenator-inputl* id (max:make-number idsinkl))

    (maxdb:type-add-value *sink-type* :id idsinkr)
    (maxdb:function-update *concatenator-inputr* id (max:make-number idsinkr))

    (maxdb:type-add-value *source-type* :id idsource)
    (maxdb:function-update *concatenator-output* id (max:make-number idsource))

;    (maxvmm:

    instance))


;(defmethod cat-inputl-sink ((cobj glcat))
;  (maxdb:db-apply *db* *concatenator-inputl* (glcat-id cobj)))

;(defmethod cat-inputr-sink ((cobj glcat))
;  (maxdb:db-apply *db* *concatenator-inputr* (glcat-id cobj)))

(defmethod obj-source ((cobj glcat))
;  (maxdb:db-apply *db* *concatenator-output* (glcat-id cobj)))
  (glcat-output cobj))

; return first available sink.
(defmethod obj-sink ((cobj glcat))
  (format t "obj-sink called on a glcat~%")
  (let* ((sinkl (glcat-inputl cobj))
         (sinkr (glcat-inputr cobj))
         (pipel (maxdb:relation-queryl *pipe* (max:make-number (sink-id sinkl))))
         (piper (maxdb:relation-queryl *pipe* (max:make-number (sink-id sinkr)))))
    (cond ((null pipel)
           sinkl)
          ((null piper)
           sinkr)
          (t nil))))

(defmethod obj-x ((cobj glcat))
  (maxdb:db-apply *db* *concatenator-x* (glcat-id cobj)))
(defmethod obj-x-update ((cobj glcat) x)
  (maxdb:function-update *concatenator-x* (glcat-id cobj) (max:make-number x)))
;(defsetf obj-x obj-x-update);already done for textobj

(defmethod obj-y ((cobj glcat))
  (maxdb:db-apply *db* *concatenator-y* (glcat-id cobj)))
(defmethod obj-y-update ((cobj glcat) y)
  (maxdb:function-update *concatenator-y* (glcat-id cobj) (max:make-number y)))
;(defsetf obj-y obj-y-update)

(defmethod obj-z ((cobj glcat))
  (maxdb:db-apply *db* *concatenator-z* (glcat-id cobj)))
(defmethod obj-z-update ((cobj glcat) z)
  (maxdb:function-update *concatenator-z* (glcat-id cobj) (max:make-number z)))
;(defsetf obj-z obj-z-update)

(defmethod obj-delete ((cobj glcat))
  (let* ((id (glcat-id cobj))
         (idsinkl (maxdb:db-apply *db* *concatenator-inputl* id))
         (idsinkr (maxdb:db-apply *db* *concatenator-inputr* id))
         (idsource (maxdb:db-apply *db* *concatenator-output* id)))

    ; remove all function mappings referencing this concatenator
    (maxdb:function-remove-mapping *concatenator-inputl* id)
    (maxdb:function-remove-mapping *concatenator-inputr* id)
    (maxdb:function-remove-mapping *concatenator-output* id)
    (maxdb:function-remove-mapping *concatenator-scene* id)
    (maxdb:function-remove-mapping *concatenator-x* id)
    (maxdb:function-remove-mapping *concatenator-y* id)
    (maxdb:function-remove-mapping *concatenator-z* id)

    ; remove the concatenator type id
    (maxdb:type-remove-value *concatenator-type* :id id)

    ; remove linked source and sink type ids
    (maxdb:type-remove-value *sink-type* :id idsinkl)
    (maxdb:type-remove-value *sink-type* :id idsinkr)
    (maxdb:type-remove-value *source-type* :id idsource)))
