(require :maxdb)

(defparameter *dbh* (make-instance 'maxdb:file-database :filename "db.max"))
(defparameter *gltexttype* (maxdb:make-type *dbh* :id 'gltext))
(defparameter *gltext-string* (maxdb:make-function *dbh* *gltexttype* maxdb:*string-type* :id 'gltext-string))
(defparameter *gltext-x* (maxdb:make-function *dbh* *gltexttype* maxdb:*number-type* :id 'gltext-x))
(defparameter *gltext-y* (maxdb:make-function *dbh* *gltexttype* maxdb:*number-type* :id 'gltext-y))
(defparameter *gltext-z* (maxdb:make-function *dbh* *gltexttype* maxdb:*number-type* :id 'gltext-z))
(defparameter *n* 0)
(defparameter *nscene* 0)
(defparameter *scene-type* (maxdb:make-type *dbh* :id 'scene))
(defparameter *gltext-scene* (maxdb:make-function *dbh* *gltexttype* *scene-type* :id 'gltext-scene))
(defparameter *scene-name* (maxdb:make-function *dbh* *scene-type* maxdb:*string-type* :id 'scene-name))

(defparameter *scene* nil)

(defclass scene () nil); placeholder since datastore calls make-instance on it

(defstruct pos (x 0f0) (y 0f0) (z 0f0))
(defstruct color (r 0f0) (g 0f0) (b 0f0) (a 1f0))

(defclass gltext () 
  ((text :initarg :text :accessor text)
   (pos :initarg :offset :accessor pos)))

(defclass glshape ()
  ((pos :initarg :offset :accessor pos)
   (color :initarg :color :accessor color)
   (primitive :initarg :primitive :accessor primitive)
   (vertices :initarg :vertices :accessor vertices)))

(defmethod scene-add-shape ((s scene) (shape glshape))
  (format t "Shapes currently not supported, not converting.~%"))

(defmethod scene-add-text ((s scene) (text gltext))
  (let ((p (pos text)))
    (format t "text: ~A~%coords: <~a,~a,~a>~%" (text text) (pos-x p) (pos-y p) (pos-z p))
    (maxdb:type-add-value *gltexttype* :id *n*)
    (maxdb:function-update *gltext-scene* *n* (max:make-number *nscene*))
    (maxdb:function-update *gltext-string* *n* (max:make-string-address (text text)))
    (maxdb:function-update *gltext-x* *n* (max:make-number (pos-x p)))
    (maxdb:function-update *gltext-y* *n* (max:make-number (pos-y p)))
    (maxdb:function-update *gltext-z* *n* (max:make-number (pos-z p)))
    (incf *n*)))


(defun convert (sourcefile)
  (load sourcefile))      

(defun convert-all ()
  (let ((dirlist (directory "ds-*.lisp")))
    (dolist (path dirlist)
      (let ((scene-name (subseq (pathname-name path) 3)))
        (maxdb:type-add-value *scene-type* :id *nscene*)
        (maxdb:function-update *scene-name* *nscene* (max:make-string-address scene-name))
        (convert path)
        (incf *nscene*)))))
